// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "InternalSessionFactory.h"
#include "InternalDecoderSession.h"
#include "InternalScalerSession.h"
#include "InternalFilterSession.h"
#include "InternalDmaFilterSession.h"
#include "InternalMioFilterSession.h"
#include "InternalEncoderSession.h"

#include "InternalSideData.h"
#include "InternalFrame.h"
#include "InternalDataBuffer.h"

#include <cstring>
#include <cstdarg>
#include <time.h>
#include <chrono>
#include <thread>

namespace xma {

#define XMA_SESSION_FACTORY "SessionFactory"

///////////////////////////////////////////////////////////////////////////////
XmaAPIVersion SessionFactory::GetLatestAPIVersion() {
  return XMA_API_VERSION_LATEST;
}

///////////////////////////////////////////////////////////////////////////////
SessionFactoryPtr SessionFactory::MakeSessionFactory(LogPtr log, XmaInitParameter* param) {
  log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_FACTORY, "SessionFactory::MakeSessionFactory()");
  SessionFactoryPtr ret = std::make_shared<InternalSessionFactory>(log, param, false);
  log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_FACTORY, "SessionFactory::MakeSessionFactory() - Returning");
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
InternalSessionFactory::InternalSessionFactory(LogPtr log, XmaInitParameter* param, bool is_c) : m_log(log), m_is_c(is_c) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory()");
  sprintf(m_device, "/dev/ama_transcoder%u", param->device);
  m_dev_ctx = Ama::CreateHwContext(m_device);
  if (!m_dev_ctx) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() Failed to create VPI device");
  }

  for (unsigned i = 0; i < param->param_cnt; ++i) {
    char text[64];
    switch (param->params[i].type) {
    case XMA_STRING:
      if (vpi_device_set_param(m_dev_ctx->GetDevHandle(), param->params[i].name, static_cast<char*>(param->params[i].value)) < 0) {
        m_dev_ctx.reset();
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() Unable to set device parameter %s = \"%s\"",
            param->params[i].name, static_cast<char*>(param->params[i].value));
      }
      break;
    case XMA_INT32:
      sprintf(text, "%d", *static_cast<int32_t*>(param->params[i].value));
      if (vpi_device_set_param(m_dev_ctx->GetDevHandle(), param->params[i].name, text) < 0) {
        m_dev_ctx.reset();
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() Unable to set device parameter %s = \"%s\"",
            param->params[i].name, text);
      }
      break;
    case XMA_UINT32:
      if (strcmp(param->params[i].name, XMA_API_VERSION) == 0) {
        m_api_version = static_cast<XmaAPIVersion>(*static_cast<uint32_t*>(param->params[i].value));
      } else {
        sprintf(text, "%u", *static_cast<uint32_t*>(param->params[i].value));
        if (vpi_device_set_param(m_dev_ctx->GetDevHandle(), param->params[i].name, text) < 0) {
          m_dev_ctx.reset();
          THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY,
              "InternalSessionFactory::InternalSessionFactory() Unable to set device parameter %s = \"%s\"", param->params[i].name, text);
        }
      }
      break;
    case XMA_INT64:
      sprintf(text, "%ld", *static_cast<int64_t*>(param->params[i].value));
      if (vpi_device_set_param(m_dev_ctx->GetDevHandle(), param->params[i].name, text) < 0) {
        m_dev_ctx.reset();
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() Unable to set device parameter %s = \"%s\"",
            param->params[i].name, text);
      }
      break;
    case XMA_UINT64:
      sprintf(text, "%lu", *static_cast<uint64_t*>(param->params[i].value));
      if (vpi_device_set_param(m_dev_ctx->GetDevHandle(), param->params[i].name, text) < 0) {
        m_dev_ctx.reset();
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() Unable to set device parameter %s = \"%s\"",
            param->params[i].name, text);
      }
      break;
    case XMA_DOUBLE:
      sprintf(text, "%f", *static_cast<double*>(param->params[i].value));
      if (vpi_device_set_param(m_dev_ctx->GetDevHandle(), param->params[i].name, text) < 0) {
        m_dev_ctx.reset();
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() Unable to set device parameter %s = \"%s\"",
            param->params[i].name, text);
      }
      break;
    default:
      m_dev_ctx.reset();
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() Unknown parameter data type %d", param->params[i].type);
    }
  }

  switch (m_api_version) {
  case XMA_API_VERSION_1_0:
  case XMA_API_VERSION_1_1:
  case XMA_API_VERSION_1_1_2:
    break;
  default:
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() Unsupported API version %u!", m_api_version);
  }

  m_mem_ctx = m_dev_ctx->GetMemCtx();
  if (!m_mem_ctx) {
    m_dev_ctx.reset();
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() Failed to create VPI memory!");
  }

  try {
    m_buffer_pool[0] = std::make_shared<BufferPool>(m_mem_ctx, m_log, MAX_POOL_ITEM_NUM, 0, 0);
  } catch (...) {
    m_dev_ctx.reset();
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() Failed to create buffer pool!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::InternalSessionFactory() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalSessionFactory::~InternalSessionFactory() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::~InternalSessionFactory()");
#ifdef XMA_CHECK_MEM
  if (!m_frames.empty()) {
    m_log->LogMsg(XMA_ERROR_LOG, XMA_SESSION_FACTORY, "XMA Memory check Failed! XmaFrame count(%d) != 0.", static_cast<int32_t>(m_frames.size()));
  }
  if (!m_data_buffers.empty()) {
    m_log->LogMsg(XMA_ERROR_LOG, XMA_SESSION_FACTORY, "XMA Memory check Failed! XmaDataBuffer count(%d) != 0.", static_cast<int32_t>(m_data_buffers.size()));
  }
  if (!m_side_data.empty()) {
    m_log->LogMsg(XMA_ERROR_LOG, XMA_SESSION_FACTORY, "XMA Memory check Failed! XmaSideDataHandle count(%d) != 0.", static_cast<int32_t>(m_side_data.size()));
  }
  if ((m_frames.empty()) && (m_data_buffers.empty()) && (m_side_data.empty())) {
    m_log->LogMsg(XMA_ERROR_LOG, XMA_SESSION_FACTORY, "XMA Memory check Succeeded!");
  }
#endif

  m_buffer_pool.clear();
  m_side_data_pool[0].clear();
  m_side_data_pool[1].clear();
  m_side_data_pool[2].clear();
  m_dev_ctx.reset();

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::~InternalSessionFactory() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
DecoderSessionPtr InternalSessionFactory::MakeDecoderSession(XmaDecoderProperties* props) {
  return std::make_unique<InternalDecoderSession>(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), props);
}

///////////////////////////////////////////////////////////////////////////////
ScalerSessionPtr InternalSessionFactory::MakeScalerSession(XmaScalerProperties* props) {
  return std::make_unique<InternalScalerSession>(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), props);
}

///////////////////////////////////////////////////////////////////////////////
FilterSessionPtr InternalSessionFactory::MakeFilterSession(XmaFilterProperties* props) {
  if ((props->hwfilter_type == XMA_UPLOAD_FILTER_TYPE) || (props->hwfilter_type == XMA_DOWNLOAD_FILTER_TYPE)) {
    return std::make_unique<InternalDmaFilterSession>(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), props);
  } else {
    return std::make_unique<InternalFilterSession>(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), props);
  }
}

///////////////////////////////////////////////////////////////////////////////
MioFilterSessionPtr InternalSessionFactory::MakeMioFilterSession(XmaMioFilterProperties* props) {
  return std::make_unique<InternalMioFilterSession>(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), props);
}

///////////////////////////////////////////////////////////////////////////////
EncoderSessionPtr InternalSessionFactory::MakeEncoderSession(XmaEncoderProperties* props) {
  return std::make_unique<InternalEncoderSession>(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), props);
}

///////////////////////////////////////////////////////////////////////////////
SideDataPtr InternalSessionFactory::AllocSideData(enum XmaFrameSideDataType type, XmaBufferType buffer_type, uint64_t size) {
  if (m_api_version < XMA_API_VERSION_1_1) {
    if (type >= XMA_FRAME_SIDE_DATA_ML_ROI) {
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::AllocSideData() Unsupported feature in this version of the API!");
    }
  } else {
    if ((type == XMA_FRAME_SIDE_DATA_HDR10_PARAMS) || ((type > XMA_FRAME_SIDE_DATA_SEI_USER_UNREGISTERED) && (type < XMA_FRAME_SIDE_DATA_USER_DEFINED_START))) {
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::AllocSideData() Unsupported feature in this version of the API!");
    }
  }
  IntSideDataPtr sd = InternalSideData::AllocSideData(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), m_log, type, buffer_type, size);
  return sd;
}

///////////////////////////////////////////////////////////////////////////////
FramePropertiesPtr InternalSessionFactory::AllocFrameProperties(enum XmaFormatType format, enum XmaFormatType sw_format, int32_t width, int32_t height, uint32_t flags) {
  return std::make_shared<InternalFrameProperties>(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), m_log, format, sw_format, width, height, flags);
}

///////////////////////////////////////////////////////////////////////////////
FramePtr InternalSessionFactory::AllocFrame(FramePropertiesPtr props, XmaFraction frame_rate, int64_t pts, bool is_idr) {
  if (props->GetFormat() == XMA_VPE_FMT_TYPE) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::AllocFrame() Unable to allocate a device frame!");
  }
  buffer_t buffer = GetBuffer(InternalFrameProperties::GetBufferSize(props));
  return std::make_shared<InternalFrame>(
      std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), m_log, props, false, std::move(buffer), frame_rate, pts, is_idr);
}

///////////////////////////////////////////////////////////////////////////////
FramePtr InternalSessionFactory::CloneFrame(
    FramePropertiesPtr props, uint8_t** frame_data, std::function<void(uint8_t**)> free_callback, XmaFraction frame_rate, int64_t pts, bool is_idr) {
  return std::make_shared<InternalFrame>(
      std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), m_log, props, frame_data, free_callback, frame_rate, pts, is_idr);
}

///////////////////////////////////////////////////////////////////////////////
FramePtr InternalSessionFactory::CloneFrame(FramePtr frame, XmaFraction frame_rate, int64_t pts, bool is_idr) {
  return std::make_shared<InternalFrame>(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), m_log, frame, frame_rate, pts, is_idr);
}

///////////////////////////////////////////////////////////////////////////////
DataBufferPtr InternalSessionFactory::AllocDataBuffer(size_t size, bool dummy, int64_t pts, int64_t poc) {
  return std::make_shared<InternalDataBuffer>(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), m_log, size, dummy, pts, poc);
}

///////////////////////////////////////////////////////////////////////////////
DataBufferPtr InternalSessionFactory::CloneDataBuffer(uint8_t* data, size_t size, std::function<void(uint8_t*)> free_callback, int64_t pts, int64_t poc) {
  return std::make_shared<InternalDataBuffer>(std::static_pointer_cast<InternalSessionFactory>(shared_from_this()), m_log, data, size, free_callback, pts, poc);
}

///////////////////////////////////////////////////////////////////////////////
bool InternalSessionFactory::IsDemoEnabled() const {
  char* value = getenv("AMA_DEMO");
  if (value && (strcmp(value, "1") == 0)) {
    return true;
  }
  return false;
}

///////////////////////////////////////////////////////////////////////////////
LogPtr InternalSessionFactory::GetLog() {
  return m_log;
}

///////////////////////////////////////////////////////////////////////////////
VpiDevHandle InternalSessionFactory::GetDevHandle() const {
  return m_dev_ctx->GetDevHandle();
}

///////////////////////////////////////////////////////////////////////////////
VpiMemHandle InternalSessionFactory::GetMemHandle() const {
  return m_mem_ctx;
}

///////////////////////////////////////////////////////////////////////////////
VpiMemHandle InternalSessionFactory::CreateMemHandle() const {
  return m_dev_ctx->CreateMemCtx();
}

///////////////////////////////////////////////////////////////////////////////
side_data_buffer_t InternalSessionFactory::GetSideDataBuffer(XmaFrameSideDataType type, uint32_t size, XmaBufferType buffer_type) {
  if (buffer_type == NO_BUFFER) {
    return nullptr;
  } else if ((buffer_type < XMA_HOST_BUFFER_TYPE) || (buffer_type > XMA_DEVICE_ONLY_BUFFER_TYPE)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_FACTORY, "InternalSessionFactory::GetSideDataBuffer() Invalid buffer type!");
  }
  SDBufferPoolPtr sd_pool = m_side_data_pool[buffer_type - 1][type];
  if (!sd_pool) {
    int flags = 0;
    switch (buffer_type) {
    case NO_BUFFER:
      return nullptr;
    case XMA_HOST_BUFFER_TYPE:
      flags = RC_SIDE_EN;
      break;
    case XMA_DEVICE_ONLY_BUFFER_TYPE:
      flags = EP_SIDE_EN;
      break;
    case XMA_DEVICE_BUFFER_TYPE:
      flags = RC_SIDE_EN | EP_SIDE_EN;
      break;
    }
    sd_pool                                 = std::make_shared<SDBufferPool>(m_mem_ctx, m_log, type, SD_BUFFER_POOL_AUTO_EXTEND_DEFAULT_SIZE, size, flags);
    m_side_data_pool[buffer_type - 1][type] = sd_pool;
  }
  return sd_pool->GetBuffer(size);
}

///////////////////////////////////////////////////////////////////////////////
buffer_t InternalSessionFactory::GetBuffer(uint32_t size) {
  BufferPoolPtr buffer_pool = m_buffer_pool[size];
  if (!buffer_pool) {
    buffer_pool         = std::make_shared<BufferPool>(m_mem_ctx, m_log, BUFFER_POOL_AUTO_EXTEND_DEFAULT_SIZE, size, RC_SIDE_EN, true);
    m_buffer_pool[size] = buffer_pool;
  }
  return buffer_pool->GetBuffer();
}

} // namespace xma
