// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

#include <map>
#include <functional>
#include <mutex>
#include "xmacpp/DecoderSession.h"
#include "SessionBase.h"
#include "InternalFrame.h"
#include "InternalDataBuffer.h"

namespace xma {
//////////////////////////////////////////////////////////////////////////////
#define XMA_DEFAULT_PKT_POOL_SIZE 128
#define XMA_DEFAULT_PKT_SIZE 4096

class InternalDecoderSession final : virtual public DecoderSession, virtual public SessionBase {
public:
  InternalDecoderSession(IntSessionFactoryPtr sf, XmaDecoderProperties* props);
  virtual ~InternalDecoderSession();

  virtual FramePropertiesPtr GetProps() const override;

  virtual int32_t SetLog(LogPtr log) override;

  virtual int32_t Enqueue(queue_t& data) override;
  virtual int32_t Dequeue(queue_t& data) override;

private:
  int32_t     PushPacket(DataBufferPtr buf);
  IntFramePtr GetFrame();
  void        Flush();

private:
  static void Thread(InternalDecoderSession* session);

protected:
  using lock_t           = std::unique_lock<std::mutex>;
  using vpi_mem_handle_t = std::unique_ptr<void, std::function<void(VpiMemHandle mem_ctx)>>;
  using dec_handle_t     = std::unique_ptr<void, std::function<void(void*)>>;
  using pkt_handle_t     = std::unique_ptr<VpiPkt, std::function<void(VpiPkt*)>>;
  using pool_handle_t    = std::unique_ptr<void, std::function<void(void*)>>;
  using pic_handle_t     = std::unique_ptr<VpiFrm, std::function<void(VpiFrm*)>>;
  using pts_map_t        = std::map<int64_t, int64_t>;
  using buf_map_t        = std::map<DataBuffer*, std::pair<int64_t, IntDataBufferPtr>>;
  mutable std::mutex                 m_mutex;
  vpi_mem_handle_t                   m_mem_ctx;
  VpiDecInitParams                   m_params;
  dec_handle_t                       m_handle;
  bool                               m_use_thread = false;
  std::thread                        m_thread;
  std::condition_variable            m_thread_enqueue_cond; // thread waits for next frame
  std::condition_variable            m_thread_dequeue_cond; // thread waits to push packet
  std::condition_variable            m_wait_enqueue_cond;   // push waits to be able to push frame
  std::condition_variable            m_wait_dequeue_cond;   // pull wait for next packet
  FramePropertiesPtr                 m_props;
  IntDataBufferPtr                   m_buf;
  IntFramePtr                        m_frame;
  pkt_handle_t                       m_pkt;
  pool_handle_t                      m_pool;
  pic_handle_t                       m_pic;
  bool                               m_resetting   = false;
  bool                               m_stop        = false;
  int32_t                            m_wait        = 0;
  bool                               m_low_latency = false;
  bool                               m_eos_pushed  = false;
  bool                               m_eos_found   = false;
  bool                               m_eos_pulled  = false;
  int32_t                            m_pool_size   = 0;
  XmaFraction                        m_framerate;
  buf_map_t                          m_buffer_map;
  pts_map_t                          m_pts_map;
  int64_t                            m_frame_sent     = 0;
  int64_t                            m_frame_recv     = 0;
  int64_t                            m_frame_enqueued = 0;
  int64_t                            m_frame_dequeued = 0;
  char                               m_instance_name[64];
  ExceptionStore                     m_exception;
  xma::DecPropChangeCallbackFunction m_prop_callback = nullptr;
  void*                              m_prop_param    = nullptr;
};

using IntDecoderSessionPtr = std::shared_ptr<InternalDecoderSession>;

}; // Namespace xma
