// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.


#include "xlnx_ml_arg_parse.h"


#define FLAG_HELP         "help"
#define FLAG_LOG_LEVEL    "log_level"
#define FLAG_LOG_LOCATION "log_location"
#define FLAG_LOG_FILE     "log_file"
#define FLAG_DEVICE_ID    "d"
#define FLAG_WIDTH        "w"
#define FLAG_HEIGHT       "h"
#define FLAG_PIX_FMT      "pix_fmt"
#define FLAG_INPUT_FILE   "i"
#define FLAG_MODEL_NAME   "model_name"
#define FLAG_MODEL_ARG    "model_arg"
#define FLAG_ML_PERF      "perf"
#define FLAG_TENSOR_DUMP  "tensor_dump"
#define FLAG_INF_REPEAT   "inference_repeat"
#define FLAG_FRAMES       "frames"
#define FLAG_FPS          "fps"
#define FLAG_OUTPUT_FILE  "o"

typedef enum {
    HELP_ARG = 0,
    LOG_LEVEL_ARG,
    LOG_LOCATION_ARG,
    LOG_FILE_ARG,
    DEVICE_ID_ARG,
    WIDTH_ARG,
    HEIGHT_ARG,
    PIX_FMT_ARG,
    INPUT_FILE_ARG,
    MODEL_NAME_ARG,
    MODEL_ARG_ARG,
    ML_PERF_ARG,
    ML_TENSORDUMP_ARG,
    ML_INFREPEAT_ARG,
    NUM_FRAMES_ARG,
    FPS_ARG,
    OUTPUT_FILE_ARG
} ml_argument_identifiers;

struct option ml_options[] = {
    {FLAG_HELP, no_argument, 0, HELP_ARG},
    {FLAG_LOG_LEVEL, required_argument, 0, LOG_LEVEL_ARG},
    {FLAG_LOG_LOCATION, required_argument, 0, LOG_LOCATION_ARG},
    {FLAG_LOG_FILE, required_argument, 0, LOG_FILE_ARG},
    {FLAG_DEVICE_ID, required_argument, 0, DEVICE_ID_ARG},
    {FLAG_WIDTH, required_argument, 0, WIDTH_ARG},
    {FLAG_HEIGHT, required_argument, 0, HEIGHT_ARG},
    {FLAG_PIX_FMT, required_argument, 0, PIX_FMT_ARG},
    {FLAG_INPUT_FILE, required_argument, 0, INPUT_FILE_ARG},
    {FLAG_MODEL_NAME, required_argument, 0, MODEL_NAME_ARG},
    {FLAG_MODEL_ARG, required_argument, 0, MODEL_ARG_ARG},
    {FLAG_ML_PERF, required_argument, 0, ML_PERF_ARG},
    {FLAG_TENSOR_DUMP, required_argument, 0, ML_TENSORDUMP_ARG},
    {FLAG_INF_REPEAT, required_argument, 0, ML_INFREPEAT_ARG},
    {FLAG_FRAMES, required_argument, 0, NUM_FRAMES_ARG},
    {FLAG_FPS, required_argument, 0, FPS_ARG},
    {FLAG_OUTPUT_FILE, required_argument, 0, OUTPUT_FILE_ARG},
    {0, 0, 0, 0}};

/**
 * xlnx_ml_get_help: Return the help message for printing.
 * @return The help message
 */
static char* xlnx_ml_get_help(char* program_name)
{
    char* help_msg = malloc(2048);
    sprintf(
        help_msg,
        "This program performs inference as per given model name.\n\n"
        "Usage:\n"
        "\t%s [options] -w <input-width> -h <input-height> -i\n"
        "\t<input-file> [ml_options] -w <output-1-width> -h\n"
        "\t<output-1-height> -o <output-1-file> -w ...\n\n"
        "Arguments:\n"
        "\t--help                     Print this message and exit\n"
        "\t-log <level>               Specify the log level\n"
        "\t-log_location <value>      Log location. 0 for none, \n"
        "\t                           1 console (default), 2 syslog, 3 file\n"
        "\t-log_file <file>           Name and path of the log file. Default\n"
        "\t                            ma35_ml_app.log\n"
        "\t-d <device-id>             Specify a device on which to run.\n"
        "\t                           Default: 0\n\n"
        "Input Arguments:\n\n"
        "\t-w <width>                 Specify the input's width\n"
        "\t-h <height>                Specify the input's height\n"
        "\t-pix_fmt <pixel-format>    Pixel format of the input file (nv12,\n"
        "\t                           xv15, yuv420p10le). Default : nv12.\n"
        "\t-i <input-file>            Input file to be used\n\n"
        "ML Arguments:\n\n"
        "\t-model_name <string>       Specify the model name for inference\n"
        "\t-model_arg <string>        Provide model arguments if any\n"
        "\t-perf <value>              Enable performance loggin, \n"
        "\t                           0 disabled (default), 1 enabled\n"
        "\t-tensor_dump <value>       Enable ML output tensor dump\n"
        "\t                           0 disabled (default), 1 enabled\n"
        "\t-inference_repeat <value>  Count of inference loop, Default : 1.\n\n"
        "Output Arguments:\n"
        "\t-frames <frame-count>      Number of frames to be processed.\n"
        "\t-o <file>                  File to which output is written.\n",
        program_name);
    return help_msg;
}

/**
 * xlnx_ml_validate_args: Check that the argument values are
 * valid ranges.
 * @param arguments: The arguments struct
 */
static void xlnx_ml_validate_args(XlnxMLArguments* arguments)
{
    if (access(arguments->input_file, F_OK) != 0) {
        fprintf(stderr, "INPUT FILE \"%s\" DOES NOT EXIST!\n",
                arguments->input_file);
        exit(XMA_APP_ERROR);
    }
    if (arguments->fps_num < 0 || arguments->fps_den <= 0) {
        fprintf(stderr, "Invalide fps values %d / %d\n", arguments->fps_num,
                arguments->fps_den);
        exit(XMA_APP_ERROR);
    }
}

/**
 * xlnx_ml_set_param_from_fraction: Set the parameter if the
 * source is of fractional format.
 * @param numerator: Where to store the numerator
 * @param denominator: Where to store the denominator
 * @param source: The source string which should be a number or fraction.
 * @return XMA_APP_SUCCESS on success, XMA_APP_ERROR on error
 */
static int xlnx_ml_set_param_from_fraction(int* numerator, int* denominator,
                                             char* source)
{
    if (!xlnx_utils_check_if_pattern_matches("^[0-9]+(\\/[0-9]+)?$", source)) {
        return XMA_APP_ERROR;
    }
    const char* frac = "/";
    char*      tok   = strtok(source, frac);
    *numerator       = atoi(tok);
    tok              = strtok(0, frac);
    if (tok == NULL) {
        return XMA_APP_ERROR;
    }
    *denominator = atoi(tok);
    return XMA_APP_SUCCESS;
}

/**
 * xlnx_ml_set_param_from_float: Set the parameter if the source
 * is a float
 * @param numerator: Where to store the numerator
 * @param denominator: Where to store the denominator
 * @param source: The source string which should be a number or fraction.
 * @return XMA_APP_SUCCESS on success, XMA_APP_ERROR on error
 */
static int xlnx_ml_set_param_from_float(int* numerator, int* denominator,
                                          char* source)
{
    if (!xlnx_utils_check_if_pattern_matches("^[0-9]+(.[0-9]+)?$", source)) {
        return XMA_APP_ERROR;
    }
    float raw_value = atof(source);
    *numerator      = (int)ceil(raw_value);
    *denominator    = 1;
    return XMA_APP_SUCCESS;
}

/**
 * xlnx_ml_set_fractional_param: Set the fractional parameter
 * from source
 * @param numerator: Where to store the numerator
 * @param denominator: Where to store the denominator
 * @param source: The source string which should be a number or fraction.
 * @param param_name: The name of the parameter to be set (For error message)
 * @return XMA_APP_SUCCESS on success, XMA_APP_ERROR on error
 */
static int xlnx_ml_set_fractional_param(int* numerator, int* denominator,
                                          char* source, char* param_name)
{
    if (xlnx_ml_set_param_from_fraction(numerator, denominator, source)) {

    } else if (xlnx_ml_set_param_from_float(numerator, denominator, source)) {

    } else {
        fprintf(stderr,
                "Unrecognized value \"%s\" for argument -%s! "
                "Make sure the value is of proper type.\n",
                source, param_name);
        return XMA_APP_ERROR;
    }
    return XMA_APP_SUCCESS;
}

/**
 * xlnx_ml_parse_commandline: Parse the commandline arguments
 * into the argument structure.
 * @param argc: The number of arguments
 * @param argv: The arguments
 * @param arguments: The arguments struct
 * @return XMA_APP_SUCCESS on success
 */
static int xlnx_ml_parse_commandline(int argc, char* argv[],
                                       XlnxMLArguments* arguments)
{
    int flag;
    int option_index;
    int ret            = XMA_APP_SUCCESS;
    int tmp            = 0;

    while (1) {
        flag = getopt_long_only(argc, argv, "", ml_options, &option_index);
        if (flag == -1) {
            break;
        }
        switch (flag) {
            case HELP_ARG:; // Need statement to follow label.
                char* help_message = xlnx_ml_get_help(argv[0]);
                fprintf(stderr, "%s\n", help_message);
                free(help_message);
                exit(XMA_APP_SUCCESS);
            case LOG_LEVEL_ARG:
                ret                  = xlnx_utils_set_int_arg(&tmp, optarg, "");
                arguments->log_level = (XmaLogLevelType)tmp;
                if (ret == XMA_APP_ERROR) {
                    ret = xlnx_utils_key_from_string(
                        param_log_level_lookup, optarg, PARAM_LOG_LEVEL_KEYS);
                    if (ret == XMA_APP_ERROR) {
                        if (optarg) {
                            fprintf(stderr, "Invalid log level %s\n", optarg);
                        } else {
                            fprintf(stderr, "Invalid log level");
                        }
                    } else {
                        arguments->log_level = ret;
                        ret                  = XMA_APP_SUCCESS;
                    }
                }
                break;
            case LOG_LOCATION_ARG:
                ret = xlnx_utils_set_int_arg(&tmp, optarg, "");
                arguments->log_location = (XmaLogType)tmp;
                if (ret == XMA_APP_ERROR) {
                    ret = xlnx_utils_key_from_string(param_log_location_lookup,
                                                     optarg,
                                                     PARAM_LOG_LOCATION_KEYS);
                    if (ret == XMA_APP_ERROR) {
                        if (optarg) {
                            fprintf(stderr, "Invalid log location value %s\n",
                                    optarg);
                        } else {
                            fprintf(stderr, "Invalid log location");
                        }
                    } else {
                        arguments->log_location = ret;
                        ret                     = XMA_APP_SUCCESS;
                    }
                }

                break;
            case LOG_FILE_ARG:
                arguments->log_file = optarg;
                ret                 = XMA_APP_SUCCESS;
                break;
            case DEVICE_ID_ARG:
                ret = xlnx_utils_set_uint_arg(&arguments->device_id, optarg,
                                              FLAG_DEVICE_ID);
                break;
            case INPUT_FILE_ARG:
                arguments->input_file = optarg;
                break;
            case MODEL_NAME_ARG:
                arguments->model_name = optarg;
                break;
            case MODEL_ARG_ARG:
                arguments->model_arg = optarg;
                break;
            case ML_PERF_ARG:
                arguments->ml_perf = optarg;
                break;
            case ML_TENSORDUMP_ARG:
                arguments->ml_tensor_dump = optarg;
                break;
            case ML_INFREPEAT_ARG:
                arguments->ml_inf_repeat = optarg;
                break;
            case WIDTH_ARG:
                ret = xlnx_utils_set_int_arg(&arguments->input_width,
                                             optarg, FLAG_WIDTH);
                break;
            case HEIGHT_ARG:
                ret = xlnx_utils_set_int_arg(&arguments->input_height,
                                             optarg, FLAG_HEIGHT);
                break;
            case FPS_ARG:
                ret = xlnx_ml_set_fractional_param(
                &arguments->fps_num, &arguments->fps_den, optarg, FLAG_FPS);
                break;
            case NUM_FRAMES_ARG:
                ret = xlnx_utils_set_size_t_arg(&arguments->num_frames, optarg,
                                                FLAG_FRAMES);
                break;
            case OUTPUT_FILE_ARG:
                arguments->output_file = optarg;
                break;
            default:
                fprintf(stderr,
                        "ERROR: Failed to properly parse commandline.");
                ret = XMA_APP_ERROR;
        }
        if (ret == XMA_APP_ERROR) {
            exit(ret);
        }
    }
    if (optind < argc) {
        fprintf(stderr, "Failed Unrecognized argument %s.\n", argv[optind]);
        exit(XMA_APP_ERROR);
    }
    return XMA_APP_SUCCESS;
}

/**
 * Set default arguments
 * @param arguments The empty argument struct
 * @return void
 */
static void xlnx_ml_set_default_arguments(XlnxMLArguments* arguments)
{
    arguments->log_level      = XMA_WARNING_LOG;
    arguments->log_location   = 1;
    arguments->log_file       = "ma35_ml_app.log";
    arguments->input_pix_fmt  = XMA_RGB24_P_FMT_TYPE;
    arguments->input_file     = "";
    arguments->model_name     = NULL;
    arguments->model_arg      = NULL;
    arguments->device_id      = DEFAULT_DEVICE_ID;
    arguments->fps_num        = 25;
    arguments->fps_den        = 1;
    arguments->num_frames     = SIZE_MAX;
    arguments->ml_perf        = "0";
    arguments->ml_tensor_dump = "0";
    arguments->ml_inf_repeat  = "1";
}

/**
 * xlnx_ml_get_arguments: Parses the commandline arguments into
 * a struct corresponding to ml arguments
 * @param argc: The number of arguments
 * @param argv: The arguments
 * @return An argument struct containing ml arguments.
 */
XlnxMLArguments xlnx_ml_get_arguments(int argc, char* argv[])
{
    if (argc < 2) {
        char* help_message = xlnx_ml_get_help(argv[0]);
        fprintf(stderr, "%s\n", help_message);
        free(help_message);
        exit(XMA_APP_SUCCESS);
    }
    XlnxMLArguments arguments;
    memset(&arguments, 0, sizeof(arguments));

    xlnx_ml_set_default_arguments(&arguments);
    xlnx_ml_parse_commandline(argc, argv, &arguments);

    xlnx_ml_validate_args(&arguments);

    return arguments;
}
